local BasePlugin = require "kong.plugins.base_plugin"
local resty_http = require 'resty.http'
local cjson = require "cjson"
local utility = require "kong.plugins.traceable.utility"

local TraceableHandler = BasePlugin:extend()

-- Note: when we release the the version below is replaced
-- by the semantic release version
TraceableHandler.VERSION = "1.0.13"

TraceableHandler.PRIORITY = 10

function onExtCapFailure(config)
  if config.allow_on_failure then
    return
  else
    return kong.response.error(403)
  end
end

function copyTraceContextHeaders(contextHeaders, headers)
  for k, v in pairs(contextHeaders) do
    if utility.isTracePropagationHeader(k) then
      headers[k] = v
    end
  end
end

function processHeaders(headers, rpcHeaders)
  local result = {}
  for k, v in pairs(headers) do
    if type(v) == "table" then
      result[k] = utility.flattenHeaderArray(v)
    else
      if utility.isTracePropagationHeader(k) then
        rpcHeaders[k] = v
      else
        result[k] = v
      end
    end
  end
  return result
end

function makeReqCapRequestBody(reqHeaders)
  local body = ""
  if ngx.req.http_version() < 2 then
    body = kong.request.get_raw_body()
  end
  return cjson.encode({
    request = {
      timestamp = utility.toProtoTimestamp(ngx.req.start_time()),
      method = kong.request.get_method(),
      headers = reqHeaders,
      scheme = kong.request.get_scheme(),
      path = kong.request.get_path_with_query(),
      host = kong.request.get_host(),
      body = ngx.encode_base64(body),
      source_address = kong.client.get_ip(),
      source_port = kong.client.get_port(),
    }
  })
end

function makeResCapRequestBody(resHeaders)
  local responseTable = getResponseTable()
  local body = table.concat(responseTable, "")
  return cjson.encode({
    response = {
      headers = resHeaders,
      body = ngx.encode_base64(body),
    }
  })
end

function sendResCapRequest(premature, config, request)
  local client = resty_http.new()
  local res, err = client:request_uri(config.ext_cap_endpoint .. "/ext_cap/v1/res_cap", request)
end

function getResponseTable()
  local plugin_ctx = kong.ctx.plugin
  if not plugin_ctx.response_table then
    plugin_ctx.response_table = {}
  end
  return plugin_ctx.response_table
end

function TraceableHandler:new()
  TraceableHandler.super.new(self, "traceable-plugin")
end

function TraceableHandler:access(config)
  TraceableHandler.super.access(self, config)

  local client = resty_http.new()
  local headers = {
    ["content-type"] = "application/json",
    ["traceableai.module.name"] = "kong",
    ["traceableai.module.version"] = TraceableHandler.VERSION,
    ["traceableai.skip_n_parents"] = "0",
  }
  local reqHeaders = processHeaders(kong.request.get_headers(), headers)
  local body = makeReqCapRequestBody(reqHeaders)
  local res, err = client:request_uri(config.ext_cap_endpoint .. "/ext_cap/v1/req_cap", {
    method = "POST",
    headers = headers,
    body = body,
  })
  if not res then
    ngx.log(ngx.ERR, "ext_cap: req_cap failed: ", err)
    return onExtCapFailure(config)
  end
  if res.status ~= 200 then
    ngx.log(ngx.ERR, "ext_cap: req_cap failed: ",res.status)
    return onExtCapFailure(config)
  end
  local authorization = cjson.decode(res.body)
  if authorization.allowRequest then
    return
  end

  return kong.response.error(authorization.statusOnDeny, authorization.messageOnDeny)
end

-- Kong 1.4 doesn't support accessing the response body through its plugin api, so
-- we add this body filter to capture it.
--
-- We might consider dropping this code and using
--   https://docs.konghq.com/gateway-oss/2.4.x/pdk/kong.service.response/#kongserviceresponseget_raw_body
-- when Kong 1.4 support is no longer necessary.
function TraceableHandler:body_filter(config)
  TraceableHandler.super.body_filter(self, config)
  local responseTable = getResponseTable()
  local chunk = ngx.arg[1]
  if chunk then
    table.insert(responseTable, chunk)
  end
end

function TraceableHandler:log(config)
  TraceableHandler.super.log(self, config)

  local headers = {
    ["content-type"] = "application/json",
    ["traceableai.module.name"] = "kong",
    ["traceableai.module.version"] = TraceableHandler.VERSION,
    ["traceableai.skip_n_parents"] = "0",
  }
  copyTraceContextHeaders(kong.request.get_headers(), headers)
  local resHeaders = processHeaders(kong.response.get_headers(), headers)
  local body = makeResCapRequestBody(resHeaders)

  local request = {
    method = "POST",
    headers = headers,
    body = body,
  }
  -- We use ngx.timer.at to work around the limitation of sockets not being available in the log
  -- phase.
  --
  -- See https://github.com/openresty/lua-nginx-module#cosockets-not-available-everywhere
  ngx.timer.at(0, sendResCapRequest, config, request)
end

return TraceableHandler
